import numpy as np
from .cell import Cell


class Manifold:
    """

    """

    def __init__(self, gain, touch):
        """

        """

        # Check that gain is of type list or type numpy.ndarray
        if not isinstance(gain, (list, np.ndarray)):
            raise TypeError(f"Gain ({type(gain)}) must be of type list or type numpy.ndarray")

        # Check that touch is of type list or type numpy.ndarray
        if not isinstance(touch, (list, np.ndarray)):
            raise TypeError(f"Touch ({type(touch)}) must be of type list or type numpy.ndarray")

        # Convert a gain list to a numpy.ndarray
        if isinstance(gain, list):
            gain = np.array(gain)

        # Convert a touch list to a numpy.ndarray
        if isinstance(touch, list):
            touch = np.array(touch)

        # Assert that gain and touch have an equal number of dimensions except for the last dimension
        assert gain.shape[:-1] == touch.shape[
                                  :-1], f"Gain and touch must have the same dimensions up to last: {gain.shape[:-1]}," \
                                        f" {touch.shape[:-1]} "

        # Check that touch is of type numpy.ndarray
        assert gain.shape[-1] >= len(gain.shape[:-1]), f"Last gain dimension must have equal or more elements than " \
                                                       f"previous dimensions: "

        # Check that touch is of type numpy.ndarray
        assert touch.shape[-1] >= len(touch.shape[:-1]), f"Touch must have "

        self.__gain = gain  # Set the gain
        self.__touch = touch  # Set the touch

        self.__create_manifold_cells()  # Create the cell matrix

        self.__checked = np.zeros(self.cells.shape, dtype=bool)  # Create a NumPy array of booleans to keep track of
        # cells that have been evaluated

    @property
    def gain(self):
        return self.__gain

    @property
    def touch(self):
        return self.__touch

    @property
    def cells(self):
        return self.__cells

    def __create_manifold_cells(self):
        """
        Creates a matrix of cell objects.


        """
        cells_shape = tuple(i - 1 for i in self.gain.shape[:-1])  # Construct a tuple based on the shape of the data
        self.__cells = np.empty(cells_shape, dtype=object)  # Create an empty numpy array of objects

        for cell_idx in range(self.cells.size):
            cell_idx_tuple = np.unravel_index(cell_idx, self.cells.shape)  #
            gain_idx_str = ", ".join([f"{i}:{i + 2}" for i in np.unravel_index(cell_idx, self.__cells.shape)])  #
            self.__cells[cell_idx_tuple] = eval(f"Cell(self.gain[{gain_idx_str}], self.touch[{gain_idx_str}])")  #

    def offset_cell(self, cell_idx_tuple, offset_tuple):
        out = []
        for s, i, t in zip(self.cells.shape, cell_idx_tuple, offset_tuple):
            tmp_s = i + t
            out.append(s - 1 if (tmp_s >= s) else 0 if (tmp_s < 0) else tmp_s)

        return tuple(out)

    def map_vec(self, vec, P, Q):

        # Check that vec is array-like
        if not isinstance(vec, (list, np.ndarray)):
            raise TypeError(f"Vector ({type(vec)}) must be of type list or type numpy.ndarray")

        # Convert a gain list to a numpy.ndarray
        if isinstance(vec, list):
            vec = np.array(vec)

        # Reshape vec if it is 1-dimensional
        if len(vec.shape) == 1:
            vec = vec[np.newaxis, :]

        #
        assert len(vec.shape) == 2, f"Vec must be 2-dimensional: {vec.shape}."

        # Check that vec dimension 1 is the same length as the last gain dimension
        assert vec.shape[-1] == getattr(self, P).shape[
            -1], f"Vec must have the same last dimension length as {P}: {vec.shape[-1]} : {getattr(self, P).shape[-1]}."

        self.__checked = self.__checked * False  # Clear the checked array

        #
        res = []

        #
        for pos, v in enumerate(vec):
            self.__checked = self.__checked * False  # Clear the checked array

            dist = np.zeros((self.cells.size, getattr(self, P).shape[
                -1]))  # Create a NumPy array to store the distance between the particle and the centroid of each cell

            for cell_idx in range(self.cells.size):
                cell_idx_tuple = np.unravel_index(cell_idx, self.cells.shape)  # Get the
                dist[cell_idx, :] = getattr(self.cells[cell_idx_tuple], f"{P}_centroid")  #

            min_cell_dist_idx = np.argmin(
                np.sqrt(((dist - v) ** 2).sum(axis=-1)))  # Get the index of the minimum distance
            min_cell_dist_idx_tuple = np.unravel_index(min_cell_dist_idx, self.cells.shape)  #

            nmo = Cell.newtons_method(
                self.cells[min_cell_dist_idx_tuple],
                getattr(self.cells[min_cell_dist_idx_tuple], f"{P}"),
                v
            )

            res.append(
                Cell.regress(
                    getattr(nmo.cell, f"{Q}"),
                    nmo.constrained_history[-1, :]
                )
            )

            """
            while True: # Returns eventually
                nmo = Cell.newtons_method(
                    self.cells[min_cell_dist_idx_tuple],
                    getattr(self.cells[min_cell_dist_idx_tuple], f"{P}"),
                    v
                )

                self.__checked[min_cell_dist_idx_tuple] = True
                min_cell_dist_idx_tuple = self.offset_cell(min_cell_dist_idx_tuple, nmo.offset)

                if self.__checked[min_cell_dist_idx_tuple] or any(nmo.offset):
                    res.append(
                        Cell.regress(
                            getattr(nmo.cell, f"{Q}"),
                            nmo.constrained_history[-1, :]
                        )
                    )
                    print(f"vector: {pos}")
                    print(res[-1])
                    break
            """

        return res

        # Assert that the

        # 1. For each point:
        # 2. Return the closest cell to the input vector via the cells' centroid
        # 3. Return the local mapping, constrained local mapping, and offset vector of the closest cell
        # 4. Mark the cell as having been evaluated
        # 4. If the offset vector is all zero, return the transformed constrained local mapping
        # 5. If the offset vector is not all zero, apply the offset
        # 6. If the offset is beyond the manifold boundary,

        # self.cells[cell_idx_tuple].

        # if self.cells[].is_evaluated



import numpy as np


class NewtonsMethodObject:
    def __init__(self, cell, history, vertices, goal, guess_vec, tol, max_iter):
        self.__cell = cell
        self.__history = history
        self.__vertices = vertices
        self.__goal = goal
        self.__guess_vec = guess_vec
        self.__tol = tol
        self.__max_iter = max_iter

        self.__constrained_history = np.copy(self.__history)
        self.__constrained_history[self.__constrained_history > 1.0] = 1.0
        self.__constrained_history[self.__constrained_history < 0.0] = 0.0

    @property
    def cell(self):
        return self.__cell

    @property
    def history(self):
        return self.__history

    @property
    def constrained_history(self):
        return self.__constrained_history

    @property
    def offset(self):
        result = self.__history[-1, :]
        return [1 if (i > 1.0) else -1 if (i < 0.0) else 0 for i in result]

    @property
    def vertices(self):
        return self.__vertices

    @property
    def goal(self):
        return self.__goal

    @property
    def guess_vec(self):
        return self.__guess_vec

    @property
    def tol(self):
        return self.__tol

    @property
    def max_iter(self):
        return self.__max_iter



import serial
import socket


class Connect:
    def __init__(self, connections=2, data_method=None, com=None, port=None, exe_location=None):
        self.ser = None
        self.mainSocket = None

        if connections == 2 or connections == 4:
            self.connections = connections
        else:
            raise Exception('Value for connections must be either 2 or 4.')

        data_method = data_method.lower().strip()

        if data_method == 'com' or data_method == 'socket' or data_method == 'exe':
            self.data_method = data_method
            self.com = com
            self.port = port
            self.exe_location = exe_location
        else:
            raise Exception('Value for data_method must be com, socket or exe.')

        if self.data_method == 'com':
            if self.com is None or not isinstance(self.com, str):
                raise Exception('When data_method value is com... com must have a string value.')
        elif self.data_method == 'socket':
            if self.port is None or not isinstance(self.port, int):
                raise Exception('When data_method value is socket... port must have a integer value.')
        elif self.data_method == 'exe':
            if self.exe_location is None or not isinstance(self.exe_location, str):
                raise Exception('When data_method value is exe... exe_location must have a string value.')

    def init(self):
        if self.data_method == 'com':
            port = "\\\\.\\" + str(self.com)
            self.ser = serial.Serial(port, 4608000, timeout=0)
        elif self.data_method == 'socket':
            self.mainSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            remote_ip = socket.gethostbyname(socket.gethostname())
            self.mainSocket.connect((remote_ip, self.port))
        elif self.data_method == 'exe':
            return

    def read_line(self, do_check=True):
        if self.data_method == 'com':
            if (do_check and self.size() > 30) or not do_check:
                return self.ser.readline()
            else:
                return
        elif self.data_method == 'socket':
            return self.mainSocket.recv(1024)
        elif self.data_method == 'exe':
            return

    def get_cords(self, do_check=True):
        read = self.read_line(do_check)
        if read is not None:
            tmp = read.decode().split(',')

            if len(tmp) == self.connections:
                return list(map(float, tmp))
            else:
                return []
        else:
            return []

    def size(self):
        if self.data_method == 'com':
            return self.ser.inWaiting()
        elif self.data_method == 'socket':
            raise Exception("This function size is not available for this data_method")
        elif self.data_method == 'exe':
            raise Exception("This function size is not available for this data_method")

    def flush(self):
        if self.data_method == 'com':
            self.ser.flush()
        elif self.data_method == 'socket':
            raise Exception("This function flush is not available for this data_method")
        elif self.data_method == 'exe':
            raise Exception("This function flush is not available for this data_method")
